from oasislmf.pytools.common.data import aggregatevulnerability_headers, aggregatevulnerability_dtype, aggregatevulnerability_fmt


headers = aggregatevulnerability_headers
dtype = aggregatevulnerability_dtype
fmt = aggregatevulnerability_fmt
cli_support = ['bintocsv', 'csvtobin']
