from oasislmf.pytools.common.data import vulnerability_headers, vulnerability_dtype, vulnerability_fmt


headers = vulnerability_headers
dtype = vulnerability_dtype
fmt = vulnerability_fmt
cli_support = ['bintocsv', 'csvtobin']
